/*jslint nomen: true*/
/*global RSVP, SimpleQuery, ComplexQuery*/
(function (jIO, RSVP, SimpleQuery, ComplexQuery) {
  "use strict";

  // Used to distinguish between operations done within the same millisecond
  function generateUniqueTimestamp(time) {

    // XXX: replace this with UUIDStorage function call to S4() when it becomes
    // publicly accessible
    var uuid = ('0000' + Math.floor(Math.random() * 0x10000)
      .toString(16)).slice(-4),
      //timestamp = Date.now().toString();
      timestamp = time.toString();
    return timestamp + "-" + uuid;
  }

  function isTimestamp(id) {
    //A timestamp is of the form 
    //"[13 digit number]-[4 numbers/lowercase letters]"
    var re = /^[0-9]{13}-[a-z0-9]{4}$/;
    return re.test(id);
  }

  function removeOldRevisions(
    substorage,
    results,
    keepDocument
  ) {
    var index,
      promise_list = [],
      seen_dict = {},
      document,
      log,
      start_index,
      new_promise_list,
      document_id;
    for (index = 0; index < results.data.rows.length; index += 1) {
      document = results.data.rows[index];
      // Count the number of revisions of each document, and delete older
      // ones.
      if (!seen_dict.hasOwnProperty(document.value.document_id)) {
        seen_dict[document.value.document_id] = {count: 0};
      }
      log = seen_dict[document.value.document_id];
      log.count += 1;
      //log.id = docum.id;

      // Record the index of the most recent edit that is before the cutoff
      if (!log.hasOwnProperty("last_edit_before_cutoff") &&
          !keepDocument({document: document, log: log})) {
        log.last_edit_before_cutoff = index;
      }

      // Record the index of the most recent put or remove
      if ((!log.hasOwnProperty("put_or_remove")) &&
          (document.value.op === "put" || document.value.op === "remove")) {
        log.put_or_remove = index;
        log.final = index;
      }

      if ((document.op === "putAttachment" ||
          document.op === "removeAttachment") &&
          log.hasOwnProperty(document.name) &&
          !log[document.name].hasOwnProperty("put_or_remove_attachment")) {
        log[document.name].put_or_remove_attachment = index;
        log.final = index;
      }
    }
    function checkIsId(document, document_id) {
      return document.value.document_id === document_id;
    }
    function removeDocument(document) {
      return substorage.remove(document.id);
    }
    // Looping over unique documents in storage
    for (document_id in seen_dict) {
      if (seen_dict.hasOwnProperty(document_id)) {

        // Identify where to cut to avoid loss of data
        log = seen_dict[document_id];
        start_index = Math.max(log.last_edit_before_cutoff, log.final + 1);

        // Record all remove promises for this document
        new_promise_list = results.data.rows
          .slice(start_index)
          .filter(checkIsId)
          .map(removeDocument);
        promise_list = promise_list.concat(new_promise_list);
      }
    }
    return RSVP.all(promise_list);
  }

  function throwCantFindError(id) {
    throw new jIO.util.jIOError(
      "RevisionStorage: cannot find object '" + id + "'",
      404
    );
  }

  function throwRemovedError(id) {
    throw new jIO.util.jIOError(
      "RevisionStorage: cannot find object '" + id + "' (removed)",
      404
    );
  }

  /**
   * The jIO RevisionStorage extension
   *
   * @class RevisionStorage
   * @constructor
   */
  function RevisionStorage(spec) {
    this._sub_storage = jIO.createJIO(spec.sub_storage);
    if (spec.hasOwnProperty("include_revisions")) {
      this._include_revisions = spec.include_revisions;
    } else {
      this._include_revisions = false;
    }
  }

  RevisionStorage.prototype.packOldRevisions = function (save_info) {
    /**
    save_info has this form:
      {
        keep_latest_num: 10,
        keep_active_revs: timestamp
      }
      keep_latest_num = x: keep at most the x latest copies of each unique doc
      keep_active_revs = x: throw away all outdated revisions from before x
    **/
    var substorage = this._sub_storage,
      option_dict = {
        sort_on: [["timestamp", "descending"]],
        select_list: ["document", "document_id", "operation"]
      },
      keep_fixed_number = save_info.hasOwnProperty("keep_latest_number");
    return substorage.allDocs(option_dict)
      .push(function (results) {
        if (keep_fixed_number) {
          return removeOldRevisions(substorage, results, function (data) {
            return data.log.count <= save_info.keep_latest_number;
          });
        }
        return removeOldRevisions(substorage, results, function (data) {
          return data.doc.id > save_info.keep_active_revisions;
        });
      });
  };

  RevisionStorage.prototype.get = function (id_in) {

    // Query to get the last edit made to this document
    var substorage = this._sub_storage,
      document_id_query,
      metadata_query,
      option_dict;

    if (this._include_revisions) {
      document_id_query = new SimpleQuery({
        operator: "<=",
        key: "timestamp",
        value: id_in
      });
    } else {
      document_id_query = new SimpleQuery({key: "document_id", value: id_in});
    }

    // Include id_in as value in query object for safety
    metadata_query = new ComplexQuery({
      operator: "AND",
      query_list: [
        document_id_query,
        new ComplexQuery({
          operator: "OR",
          query_list: [
            new SimpleQuery({key: "op", value: "remove"}),
            new SimpleQuery({key: "op", value: "put"})
          ]
        })
      ]
    });
    option_dict = {
      query: metadata_query,
      select_list: ["op"],
      limit: [0, 1],
      sort_on: [["timestamp", "descending"]]
    };


    return substorage.allDocs(option_dict)
      .push(function (results) {
        if (results.data.total_rows > 0) {
          if (results.data.rows[0].value.op === "put") {
            return substorage.get(results.data.rows[0].id)
              .push(function (result) {
                return result.doc;
              });
          }
          throwRemovedError(id_in);
        }
        throwCantFindError(id_in);
      });
  };

  RevisionStorage.prototype.put = function (id, data) {
    var substorage = this._sub_storage,
      timestamp = generateUniqueTimestamp(Date.now()),
      metadata = {
        // XXX: remove this attribute once query can sort_on id
        timestamp: timestamp,
        doc_id: id,
        doc: data,
        op: "put"
      };

    if (this._include_revisions && isTimestamp(id)) {
      return substorage.get(id)
        .push(function (metadata) {
          metadata.timestamp = timestamp;
          metadata.doc = data;
          return substorage.put(timestamp, metadata);
        },
          function (error) {
            if (error.status_code === 404 &&
                error instanceof jIO.util.jIOError) {
              return substorage.put(timestamp, metadata);
            }
            throw error;
          });
    }
    return this._sub_storage.put(timestamp, metadata);
  };

  RevisionStorage.prototype.remove = function (id) {
    var timestamp = generateUniqueTimestamp(Date.now() - 1),
      metadata = {
        // XXX: remove this attribute once query can sort_on id
        timestamp: timestamp,
        doc_id: id,
        op: "remove"
      };
    return this._sub_storage.put(timestamp, metadata);
  };

  RevisionStorage.prototype.allAttachments = function (id) {
    var substorage = this._sub_storage,
      query_object,
      query_removed_check,
      option_dict,
      query_document_id,
      check_if_removed_option_list,
      include_revisions = this._include_revisions,
      have_seen_id = false;

    // id is a timestamp, and allAttachments will return attachment versions
    // up-to-and-including those made at time id
    if (include_revisions) {
      query_document_id = new SimpleQuery({
        operator: "<=",
        key: "timestamp",
        value: id
      });
    } else {
      query_document_id = new SimpleQuery({key: "doc_id", value: id});
      have_seen_id = true;
    }

    query_removed_check = new ComplexQuery({
      operator: "AND",
      query_list: [
        query_document_id,
        new ComplexQuery({
          operator: "OR",
          query_list: [
            new SimpleQuery({key: "op", value: "put"}),
            new SimpleQuery({key: "op", value: "remove"})
          ]
        })
      ]
    });

    query_object = new ComplexQuery({
      operator: "AND",
      query_list: [
        query_document_id,
        new ComplexQuery({
          operator: "OR",
          query_list: [
            new SimpleQuery({key: "op", value: "putAttachment"}),
            new SimpleQuery({key: "op", value: "removeAttachment"})
          ]
        })
      ]
    });

    check_if_removed_option_list = {
      query: query_removed_check,
      select_list: ["op", "timestamp"],
      limit: [0, 1],
      sort_on: [["timestamp", "descending"]]
    };
    option_dict = {
      query: query_object,
      sort_on: [["timestamp", "descending"]],
      select_list: ["op", "name"]
    };

    return this._sub_storage.allDocs(check_if_removed_option_list)
      // Check the document exists and is not removed
      .push(function (results) {
        if (results.data.total_rows > 0) {
          if (results.data.rows[0].id === id) {
            have_seen_id = true;
          }
          if (results.data.rows[0].value.op === "remove") {
            throwRemovedError(id);
          }
        } else {
          throwCantFindError(id);
        }
      })
      .push(function () {
        return substorage.allDocs(option_dict);
      })
      .push(function (results) {
        var seen_dict = {},
          attachment_list = [],
          attachment_promise_list = [],
          i,
          entry;

        // If input mapped to a real timestamp, then the first query result must
        // have the inputted id.  Otherwise, unexpected results could arise
        // by inputting nonsensical strings as id when include_revisions = true
        if (include_revisions &&
            results.data.total_rows > 0 &&
            results.data.rows[0].id !== id &&
            !have_seen_id) {
          throwCantFindError(id);
        }


        // Only return attachments if:
        // (it is the most recent revision) AND (it is a putAttachment)
        attachment_list = results.data.rows.filter(function (document) {
          if (!seen_dict.hasOwnProperty(document.value.name)) {
            var output = (document.value.op === "putAttachment");
            seen_dict[document.value.name] = {};
            return output;
          }
        });
        // Assembles object of attachment_name: attachment_object
        for (i = 0; i < attachment_list.length; i += 1) {
          entry = attachment_list[i];
          attachment_promise_list[entry.value.name] =
            substorage.getAttachment(entry.id, entry.value.name);
        }
        return RSVP.hash(attachment_promise_list);
      });
  };

  RevisionStorage.prototype.putAttachment = function (id, name, blob) {
    var timestamp = generateUniqueTimestamp(Date.now()),
      metadata = {
        // XXX: remove this attribute once query can sort_on id
        timestamp: timestamp,
        doc_id: id,
        name: name,
        op: "putAttachment"
      },
      substorage = this._sub_storage;

    if (this._include_revisions && isTimestamp(id)) {
      return substorage.get(id)
        .push(function (metadata) {
          metadata.timestamp = timestamp;
          metadata.name = name;
        },
          function (error) {
            if (!(error.status_code === 404 &&
                error instanceof jIO.util.jIOError)) {
              throw error;
            }
          });
    }
    return this._sub_storage.put(timestamp, metadata)
      .push(function () {
        return substorage.putAttachment(timestamp, name, blob);
      });
  };

  RevisionStorage.prototype.getAttachment = function (id, name) {

    // In this case, id is a timestamp, so return attachment version at that
    // time
    if (this._include_revisions) {
      return this._sub_storage.getAttachment(id, name)
        .push(undefined, function (error) {
          if (error.status_code === 404 &&
              error instanceof jIO.util.jIOError) {
            throwCantFindError(id);
          }
          throw error;
        });
    }

    // Query to get the last edit made to this document
    var substorage = this._sub_storage,

      // "doc_id: id AND
      //  (op: remove OR ((op: putAttachment OR op: removeAttachment) AND
      //  name: name))"
      metadata_query = new ComplexQuery({
        operator: "AND",
        query_list: [
          new SimpleQuery({key: "doc_id", value: id}),
          new ComplexQuery({
            operator: "OR",
            query_list: [
              new SimpleQuery({key: "op", value: "remove"}),
              new ComplexQuery({
                operator: "AND",
                query_list: [
                  new ComplexQuery({
                    operator: "OR",
                    query_list: [
                      new SimpleQuery({key: "op", value: "putAttachment"}),
                      new SimpleQuery({key: "op", value: "removeAttachment"})
                    ]
                  }),
                  new SimpleQuery({key: "name", value: name})
                ]
              })
            ]
          })
        ]
      }),
      option_dict = {
        query: metadata_query,
        sort_on: [["timestamp", "descending"]],
        limit: [0, 1],
        select_list: ["op", "name"]
      };
    return substorage.allDocs(option_dict)
      .push(function (results) {
        if (results.data.total_rows > 0) {
          // XXX: issue if attachments are put on a removed document
          if (results.data.rows[0].value.op === "remove" ||
              results.data.rows[0].value.op === "removeAttachment") {
            throwRemovedError(id);
          }
          return substorage.getAttachment(results.data.rows[0].id, name);
        }
        throwCantFindError(id);
      });
  };

  RevisionStorage.prototype.removeAttachment = function (id, name) {
    var timestamp = generateUniqueTimestamp(Date.now()),
      metadata = {
        // XXX: remove this attribute once query can sort_on id
        timestamp: timestamp,
        doc_id: id,
        name: name,
        op: "removeAttachment"
      };
    return this._sub_storage.put(timestamp, metadata);
  };

  RevisionStorage.prototype.repair = function () {
    return this._sub_storage.repair.apply(this._sub_storage, arguments);
  };

  RevisionStorage.prototype.hasCapacity = function (name) {
    return name === 'list' ||
      name === 'include' ||
      name === 'query' ||
      name === 'select' ||
      name === 'limit' ||
      name === 'sort';
  };

  RevisionStorage.prototype.buildQuery = function (option_dict) {
    // Set default values
    if (option_dict === undefined) {option_dict = {}; }
    if (option_dict.query === undefined) {option_dict.query = ""; }
    if (option_dict.sort_on === undefined) {option_dict.sort_on = []; }
    if (option_dict.select_list === undefined) {option_dict.select_list = []; }
    //option_dict.query = jIO.QueryFactory.create(option_dict.query);
    var meta_option_dict  = {
      query: "",
      sort_on: [["timestamp", "descending"]],
      select_list: ["doc", "op", "doc_id", "timestamp"]
    },
      include_revisions = this._include_revisions;

    // When include_revisions is set, the document metadata will be queried 
    // instead of the documents themselves.
    if (include_revisions) {
      meta_option_dict.query = option_dict.query;
    }

    return this._sub_storage.allDocs(meta_option_dict)
      .push(function (result_list) {
        result_list = result_list.data.rows;
        var seen_dict = {},
          documents_to_query_list,
          i;

        if (include_revisions) {

          // We only query on versions mapping to puts or putAttachments
          result_list = result_list.map(function (document, index) {
            var data_key;
            if (document.value.op === "put") {
              return document;
            }
            if (document.value.op === "remove") {
              document.value.doc = {};
              return document;
            }
            if (document.value.op === "putAttachment" ||
                document.value.op === "removeAttachment") {

              // putAttachment document does not contain doc metadata, so we
              // add it from the most recent non-removed put on same id
              document.value.doc = {};
              for (i = index + 1; i < result_list.length; i += 1) {
                if (result_list[i].value.doc_id === document.value.doc_id) {
                  if (result_list[i].value.op === "put") {
                    for (data_key in result_list[i].value.doc) {
                      if (result_list[i].value.doc.hasOwnProperty(data_key)) {
                        document.value.doc[data_key] =
                          result_list[i].value.doc[data_key];
                      }
                    }
                    return document;
                  }
                  // If most recent metadata edit before the attachment edit 
                  // was a remove, then leave doc empty
                  if (result_list[i].value.op === "remove") {
                    return document;
                  }
                }
              }
            }
            return false;
          });
        } else {

          // Only query on latest revisions of non-removed documents/attachment
          // edits
          result_list = result_list.map(function (document, index) {
            var data_key;
            if (document.value.op === "put") {
              // Mark as read and include in query
              if (!seen_dict.hasOwnProperty(document.value.doc_id)) {
                seen_dict[document.value.doc_id] = {};
                return document;
              }

            } else if (document.value.op === "remove" ||
                      document.value.op === "removeAttachment") {
              // Mark as read but do not include in query
              seen_dict[document.value.doc_id] = {};

            } else if (document.value.op === "putAttachment") {
              // If latest edit, mark as read, add documentent metadata from
              // most recent put, and add to query
              if (!seen_dict.hasOwnProperty(document.value.doc_id)) {
                seen_dict[document.value.doc_id] = {};
                document.value.doc = {};
                for (i = index + 1; i < result_list.length; i += 1) {
                  if (result_list[i].value.doc_id === document.value.doc_id) {
                    if (result_list[i].value.op === "put") {
                      for (data_key in result_list[i].value.doc) {
                        if (result_list[i].value.doc.hasOwnProperty(data_key)) {
                          document.value.doc[data_key] =
                            result_list[i].value.doc[data_key];
                        }
                      }
                      return document;
                    }
                    if (result_list[i].value.op === "remove") {
                      // If most recent edit on document was a remove before
                      // this attachment, then don't include attachment in query
                      return false;
                    }
                    document.value.doc = {};
                  }
                }
              }
            }
            return false;
          });
        }
        documents_to_query_list = result_list

          // Filter out all docs flagged as false in previous map call
          .filter(function (document) {
            return document;
          })

          // Put into correct format to be passed back to query storage
          .map(function (document) {
            if (include_revisions) {
              document.id = document.value.timestamp;
              //document.doc = document.value.doc;
            } else {
              document.id = document.value.doc_id;
              //document.doc = document.value.doc;
            }
            delete document.value.doc_id;
            delete document.value.timestamp;
            delete document.value.op;
            var temp = document.value.doc;
            document.value = {};
            for (var selection in temp) {
              if (option_dict.select_list.indexOf(selection) > -1) {
                document.value[selection] = temp[selection];
              }
            }
            //document.value = {};
            return document;
          });
        return documents_to_query_list;
      });
  };

  jIO.addStorage('revision', RevisionStorage);

}(jIO, RSVP, SimpleQuery, ComplexQuery));